from typing import Dict


class ImportStringCache(dict):
    def __missing__(self, key):
        if "." not in key:
            return __import__(key)

        module_name, class_name = key.rsplit(".", 1)

        module = __import__(module_name, {}, {}, [class_name])
        handler = getattr(module, class_name)

        # We cache a NoneType for missing imports to avoid repeated lookups
        self[key] = handler

        return handler


_cache = ImportStringCache()


def import_string(path: str):
    """
    Path must be module.path.ClassName

    >>> cls = import_string('module_x.ClassY')
    """
    result = _cache[path]
    return result


def import_submodules(context: Dict, root_module: str, path: str):
    """
    Import all submodules and register them in the ``context`` namespace.

    >>> import_submodules(locals(), __name__, __path__)
    """
    import pkgutil

    for _, module_name, _ in pkgutil.walk_packages(path, root_module + "."):
        # this causes a Runtime error with model conflicts
        # module = loader.find_module(module_name).load_module(module_name)
        module = __import__(module_name, globals(), locals(), ["__name__"])
        for k, v in vars(module).items():
            if not k.startswith("_"):
                context[k] = v
        context[module_name] = module



import asyncio

from functools import partial, wraps
from typing import Callable


def coroutine(f: Callable):
    """https://github.com/pallets/click/issues/85#issuecomment-503464628"""

    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


async def run_sync(func: Callable, *args, **kwargs):
    import anyio

    if kwargs:  # pragma: no cover
        # run_sync doesn't accept 'kwargs', so bind them in here
        func = partial(func, **kwargs)
    return await anyio.to_thread.run_sync(func, *args)



import os


def get_project_path(project_name: str) -> str:
    base_path = os.path.expanduser("~")
    if not os.access(base_path, os.W_OK):
        base_path = "/tmp"

    return os.path.join(base_path, project_name)


def get_temp_path(project_name: str) -> str:
    return os.path.join("/tmp", project_name)



from typing import Type, Union


def strtobool(val: str) -> bool:
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError("invalid truth value %r" % (val,))


def to_bool(
    value: Union[str, bool, int],
    handle_none: bool = False,
    exception: Type[Exception] = TypeError,
) -> bool:
    if isinstance(value, str):
        value = strtobool(value)

    if value in (False, 0):
        return False

    if value in (True, 1):
        return True

    if handle_none and value is None:
        return False

    raise exception("The value `{}` cannot be interpreted as boolean".format(value))



import base64


def encode(value):
    return base64.b64encode(value.encode("utf-8")).decode("utf-8")


def decode(value):
    return base64.b64decode(value).decode("utf-8")


def urlsafe_b64decode(b64string):
    if isinstance(b64string, str):
        b64string = bytes(b64string, "utf-8")
    padded = b64string + b"=" * (4 - len(b64string) % 4)
    payload = base64.urlsafe_b64decode(padded)
    try:
        return payload.decode("utf-8")
    except Exception:  # noqa
        return payload



from click import ClickException


class ClippedException(ClickException):
    def __init__(self, message=None):
        super().__init__(message)

    def __repr__(self):
        if hasattr(self, "message"):
            return self.message
        return super().__repr__()

    def __str__(self):
        if hasattr(self, "message"):
            return self.message
        return super().__str__()


class SchemaError(ClippedException):
    pass



from typing import Any, Optional
from urllib.parse import parse_qs, urlencode, urljoin, urlparse, urlunparse


def clean_verify_ssl(host: str, verify_ssl: Optional[bool] = None):
    if verify_ssl is None and "https" in host:
        return True
    return verify_ssl


def clean_host(host: str):
    return host.rstrip("/")


def absolute_uri(
    url: str, host: Optional[str] = None, protocol: Optional[str] = None
) -> Optional[str]:
    if not url:
        return None

    if not host:
        return url

    host = clean_host(host)

    if "http" in host:
        return urljoin(clean_host(host) + "/", url.lstrip("/"))

    protocol = protocol or "http"
    host = f"{protocol}://{host}"

    return urljoin(clean_host(host) + "/", url.lstrip("/"))


def add_notification_referrer_param(
    url: str, provider: str, is_absolute: bool = True
) -> Optional[Any]:
    if not is_absolute:
        url = absolute_uri(url)
    if not url:
        return None
    parsed_url = urlparse(url)
    query = parse_qs(parsed_url.query)
    query["referrer"] = provider
    url_list = list(parsed_url)
    url_list[4] = urlencode(query, doseq=True)
    return urlunparse(url_list)
